"""The setup script."""

#  Copyright (c) 2023 zfit
import os
import sys

from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))

with open(
    os.path.join(here, "requirements.txt"), encoding="utf-8"
) as requirements_file:
    requirements = requirements_file.read().splitlines()

with open(
    os.path.join(here, "requirements_dev.txt"), encoding="utf-8"
) as requirements_dev_file:
    requirements_dev = requirements_dev_file.read().splitlines()

extras_require = {}
extras_require["ipyopt"] = ["ipyopt>=0.12"]
if sys.version_info[1] < 11:
    extras_require["nlopt"] = ["nlopt>=2.7.1"]
extras_require["hs3"] = ["asdf"]
# Python 3.7 not supported anymore
extras_require["uproot"] = ["awkward-pandas"]
allreq = sum(extras_require.values(), [])

tests_require = [
    "pytest>=3.4.2",  # breaks unittests
    "pytest-runner>=2.11.1",
    "pytest-rerunfailures>=6",
    "pytest-xdist",
    "pytest-ordering",
    "pytest-randomly",
    "scikit-hep-testdata",
    "pytest-timeout>=1",
    "matplotlib",  # for plots in examples
]
extras_require["all"] = allreq
extras_require["tests-nonlinux"] = tests_require + extras_require.get("nlopt", [])
extras_require["tests"] = extras_require["tests-nonlinux"] + extras_require["ipyopt"]
extras_require["dev"] = requirements_dev + extras_require["tests"]
extras_require["dev-nonlinux"] = requirements_dev + extras_require["tests-nonlinux"]
extras_require["alldev"] = list(set(extras_require["all"] + extras_require["dev"]))
alldev_nonlinux = list(set(extras_require["all"] + extras_require["dev-nonlinux"]))
alldev_nonlinux.pop(
    alldev_nonlinux.index(extras_require["ipyopt"][0])
)  # ipyopt is not available on non linux systems
extras_require["alldev-nonlinux"] = alldev_nonlinux
alldev_windows = alldev_nonlinux.copy()
alldev_windows.pop(
    alldev_windows.index("jaxlib")
)  # not available on Windows: https://github.com/google/jax/issues/438#issuecomment-939866186
extras_require["alldev-windows"] = alldev_windows

setup(
    install_requires=requirements,
    tests_require=tests_require,
    extras_require=extras_require,
    use_scm_version=True,
)



#  Copyright (c) 2024 zfit
from __future__ import annotations

import numpy as np
import tensorflow as tf

from ..core.binnedpdf import BaseBinnedPDFV1
from ..core.space import supports
from ..util.exception import SpecificFunctionNotImplemented
from ..z import numpy as znp


class BinnedTemplatePDFV1(BaseBinnedPDFV1):
    def __init__(
        self, data, sysshape=None, extended=None, norm=None, name="BinnedTemplatePDF"
    ):
        # TODO: use scalemodifier instead
        obs = data.space
        if extended is None:
            extended = True
        if sysshape is None:
            sysshape = {}
        if sysshape is True:
            import zfit

            sysshape = {
                f"sysshape_{i}": zfit.Parameter(f"auto_sysshape_{self}_{i}", 1.0)
                for i in range(data.values().shape.num_elements())
            }
        params = {}
        params.update(sysshape)
        self._template_sysshape = sysshape
        if extended is True:
            self._automatically_extended = True
            if sysshape:
                import zfit

                def sumfunc(params):
                    values = self._data.values()
                    if sysshape := list(params.values()):
                        sysshape_flat = tf.stack(sysshape)
                        sysshape = tf.reshape(sysshape_flat, values.shape)
                        values = values * sysshape
                    return znp.sum(values)

                from zfit.core.parameter import get_auto_number

                extended = zfit.ComposedParameter(
                    f"TODO_name_selfmade_{get_auto_number()}", sumfunc, params=sysshape
                )

            else:
                extended = znp.sum(data.values())
        elif extended is not False:
            self._automatically_extended = False
        super().__init__(
            obs=obs, name=name, params=params, extended=extended, norm=norm
        )

        self._data = data

    def _ext_pdf(self, x, norm):
        counts = self._counts(x, norm)
        areas = np.prod(self._data.axes.widths, axis=0)
        density = counts / areas
        return density

    @supports(norm=False)
    def _pdf(self, x, norm):
        counts = self._rel_counts(x, norm)
        areas = np.prod(self._data.axes.widths, axis=0)
        density = counts / areas
        return density

    @supports(norm="norm")
    # @supports(norm=False)
    def _counts(self, x, norm=None):
        if not self._automatically_extended:
            raise SpecificFunctionNotImplemented
        values = self._data.values()
        if sysshape := list(self._template_sysshape.values()):
            sysshape_flat = tf.stack(sysshape)
            sysshape = tf.reshape(sysshape_flat, values.shape)
            values = values * sysshape
        return values

    @supports(norm="norm")
    def _rel_counts(self, x, norm=None):
        values = self._data.values()
        if sysshape := list(self._template_sysshape.values()):
            sysshape_flat = tf.stack(sysshape)
            sysshape = tf.reshape(sysshape_flat, values.shape)
            values = values * sysshape
        return values / znp.sum(values)


# class BinnedSystematicsPDFV1(FunctorMixin, BaseBinnedPDFV1):
#
#     def __init__(self, pdf, sysshape=None, extended=None, norm=None, name="BinnedTemplatePDF"):
#         obs = data.space
#         if sysshape is None:
#             import zfit
#             sysshape = {f'sysshape_{i}': zfit.Parameter(f'auto_sysshape_{self}_{i}', 1.) for i in
#                         range(data.values().shape.num_elements())}
#         params = {}
#         params.update(sysshape)
#         if extended is None:
#             extended = znp.sum(data.values())
#         super().__init__(obs=obs, name=name, params=params, extended=extended, norm=norm)
#
#         self._data = data
#
#     def _ext_pdf(self, x, norm):
#         counts = self._counts(x, norm)
#         areas = np.prod(self._data.axes.widths, axis=0)
#         density = counts / areas
#         return density
#
#     def _pdf(self, x, norm):
#         counts = self._counts(x, norm)
#         areas = np.prod(self._data.axes.widths, axis=0)
#         density = counts / areas
#         return density
#
#     @supports(norm='norm')
#     # @supports(norm=False)
#     def _counts(self, x, norm=None):
#
#         sysshape_flat = tf.stack([p for name, p in self.params.items() if name.startswith('sysshape')])
#         counts = self._data.values()
#         sysshape = tf.reshape(sysshape_flat, counts.shape)
#         return counts * sysshape
#
#     @supports(norm='norm')
#     def _rel_counts(self, x, norm=None):
#         sysshape_flat = tf.stack([p for name, p in self.params.items() if name.startswith('sysshape')])
#         counts = self._data.values()
#         sysshape = tf.reshape(sysshape_flat, counts.shape)
#         values = counts * sysshape
#         return values / znp.sum(values)
#
# values = self._ext_pdf(None, norm)
# areas = znp.prod(self._data.axes.widths, axis=0)
# counts = values * areas
# return counts

# class BinnedTemplatePDF(PDF):
#
#     def __init__(self, data, sysshape=None, extended=None, norm=None, label="BinnedTemplatePDF"):
#         space = data.space
#         if sysshape is None:
#             sysshape = {f'sysshape_{i}': zfit.param.Parameter(f'auto_sysshape_{self}_{i}', 1.)
#                         for i in range(data.values().shape.num_elements())}
#         var = {f'axis_{i}': axis for i, axis in enumerate(space)}
#         var.update(sysshape)
#         super().__init__(var=var, label=label, extended=extended, norm=norm)
#
#         self.sysshape = sysshape
#         self.data = data
#
#     def _ext_pdf(self, var, norm):
#         counts = self._ext_integrate(var, norm)
#         # if not isinstance(x, ZfitData):
#         #     return counts
#         areas = np.prod(self.data.axes.widths, axis=0)
#         density = counts / areas
#         return density
#
#     def _ext_integrate(self, var, norm):
#         counts = self.data.values()
#         if self.sysshape is not None:
#             sysshape_flat = tf.stack([p for name, p in self.params.items() if name.startswith('sysshape')])
#             sysshape = tf.reshape(sysshape_flat, counts.shape)
#             counts = counts * sysshape
#         if self.space == var.space and self.space.is_binned \
#                 and (not norm.space or norm.space == self.space):
#             return counts



#  Copyright (c) 2022 zfit
from __future__ import annotations

from contextlib import suppress

from zfit_interface.func import ZfitFunc
from zfit_interface.variables import ZfitParam, ZfitAxis

from zfit import Data
from zfit.core.values import ValueHolder
from zfit.util.exception import SpecificFunctionNotImplemented, WorkInProgressError


def to_value_holder(var):
    if not isinstance(var, ValueHolder):
        var = ValueHolder(var)
    return var


def to_data(value, space):
    data = Data.from_tensor(obs=space, tensor=value)  # TODO
    return data


class Func(ZfitFunc):
    def __init__(self, var, output_var=None, label=None):
        self.var = var
        self.params = {k: v for k, v in var.items() if isinstance(v, ZfitParam)}
        self.space = {k: v for k, v in var.items() if isinstance(v, ZfitAxis)}
        self.output_var = output_var
        self.label = label

    def __call__(self, var=None):
        var = to_value_holder(var)
        output = self._auto_func(var)
        if self.output_var is not None:
            output = to_data(output, self.output_var)
        return output

    def values(self, *, var, options=None):
        return self._call_values(var=var, options=options)

    def _call_values(self, var=None, options=None):
        with suppress(SpecificFunctionNotImplemented):
            return self._values(var, options=options)  # TODO: auto_value?
        return self._fallback_values(var=var, options=options)

    def _fallback_values(self, var, norm, options):
        raise WorkInProgressError

    def _values(self, var, norm, options):
        raise SpecificFunctionNotImplemented



#  Copyright (c) 2022 zfit
import functools

from tensorflow.python.util.deprecation import deprecated, deprecated_args

__all__ = "deprecated", "deprecated_args", "deprecate_norm_range"  # noqa


def deprecated_norm_range(func):
    @functools.wraps(func)
    @deprecated_args(None, "Use `norm` instead.", "norm_range")
    def wrapper(*args, norm=None, norm_range=None, **kwargs):
        if norm_range is not None:
            norm = norm_range
        try:
            return func(*args, norm=norm, **kwargs)
        except TypeError as error:
            if "unexpected keyword argument 'norm'" in str(error):
                return func(*args, norm_range=norm_range, **kwargs)
            elif "got multiple values for argument 'norm'" in str(error):
                return func(*args, **kwargs)
            else:
                raise

    return wrapper



"""Special PDFs are provided in this module.

One example is a normal function `Function` that allows to simply define a non-normalizable function.
"""

#  Copyright (c) 2023 zfit
from __future__ import annotations

import functools

import tensorflow as tf

from .functor import BaseFunctor
from ..core.basemodel import SimpleModelSubclassMixin
from ..core.basepdf import BasePDF
from ..core.space import supports
from ..util import ztyping
from ..util.exception import NormRangeNotImplemented


class SimplePDF(BasePDF):
    def __init__(self, obs, func, name="SimplePDF", **params):
        super().__init__(name=name, params=params, obs=obs)
        self._unnormalized_prob_func = self._check_input_x_function(func)

    def _unnormalized_pdf(self, x):
        try:
            return self._unnormalized_prob_func(x)
        except TypeError:
            return self._unnormalized_prob_func(self, x)

    def copy(self, **override_parameters) -> BasePDF:
        override_parameters.update(func=self._unnormalized_prob_func)
        return super().copy(**override_parameters)


class SimpleFunctorPDF(BaseFunctor, SimplePDF):
    def __init__(self, obs, pdfs, func, name="SimpleFunctorPDF", **params):
        super().__init__(obs=obs, pdfs=pdfs, func=func, name=name, **params)


def raise_error_if_norm_range(func):
    func = supports(norm=False)(func)

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except (
            NormRangeNotImplemented
        ):  # TODO: silently remove norm_range? Or loudly fail?
            raise tf.errors.InvalidArgumentError(
                "Norm_range given to Function: cannot be normalized."
            )

    return wrapped


class ZPDF(SimpleModelSubclassMixin, BasePDF):
    def __init__(self, obs: ztyping.ObsTypeInput, name: str = "ZPDF", **params):
        super().__init__(obs=obs, name=name, **params)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._check_simple_model_subclass()



#  Copyright (c) 2024 zfit

from __future__ import annotations

import numpy as np
import pydantic
import tensorflow as tf

from typing import Literal

import zfit.z.numpy as znp
from zfit import z
from ..core.basepdf import BasePDF
from ..core.serialmixin import SerializableMixin
from ..core.space import ANY_LOWER, ANY_UPPER, Space
from ..serialization import SpaceRepr, Serializer
from ..serialization.pdfrepr import BasePDFRepr
from ..util import ztyping
from ..util.ztyping import ExtendedInputType, NormInputType


def _powerlaw(x, a, k):
    return a * znp.power(x, k)


@z.function(wraps="zfit_tensor")
def crystalball_func(x, mu, sigma, alpha, n):
    t = (x - mu) / sigma * tf.sign(alpha)
    abs_alpha = znp.abs(alpha)
    a = znp.power((n / abs_alpha), n) * znp.exp(-0.5 * znp.square(alpha))
    b = (n / abs_alpha) - abs_alpha
    cond = tf.less(t, -abs_alpha)
    func = z.safe_where(
        cond,
        lambda t: _powerlaw(b - t, a, -n),
        lambda t: znp.exp(-0.5 * znp.square(t)),
        values=t,
        value_safer=lambda t: znp.ones_like(t) * (b - 2),
    )
    func = znp.maximum(func, znp.zeros_like(func))
    return func


@z.function(wraps="zfit_tensor", stateless_args=False)
def double_crystalball_func(x, mu, sigma, alphal, nl, alphar, nr):
    cond = tf.less(x, mu)

    func = tf.where(
        cond,
        crystalball_func(x, mu, sigma, alphal, nl),
        crystalball_func(x, mu, sigma, -alphar, nr),
    )

    return func


# created with the help of TensorFlow autograph used on python code converted from ShapeCB of RooFit
def crystalball_integral(limits, params, model):
    mu = params["mu"]
    sigma = params["sigma"]
    alpha = params["alpha"]
    n = params["n"]

    lower, upper = limits._rect_limits_tf

    integral = crystalball_integral_func(mu, sigma, alpha, n, lower, upper)
    return integral


@z.function(wraps="zfit_tensor")
# @tf.function  # BUG? TODO: problem with tf.function and input signature
def crystalball_integral_func(mu, sigma, alpha, n, lower, upper):
    sqrt_pi_over_two = np.sqrt(np.pi / 2)
    sqrt2 = np.sqrt(2)

    use_log = tf.less(znp.abs(n - 1.0), 1e-05)
    abs_sigma = znp.abs(sigma)
    abs_alpha = znp.abs(alpha)
    tmin = (lower - mu) / abs_sigma
    tmax = (upper - mu) / abs_sigma

    alpha_negative = tf.less(alpha, 0)
    # do not move on two lines, logic will fail...
    tmax, tmin = znp.where(alpha_negative, -tmin, tmax), znp.where(
        alpha_negative, -tmax, tmin
    )

    if_true_4 = (
        abs_sigma
        * sqrt_pi_over_two
        * (tf.math.erf(tmax / sqrt2) - tf.math.erf(tmin / sqrt2))
    )

    a = znp.power(n / abs_alpha, n) * znp.exp(-0.5 * tf.square(abs_alpha))
    b = n / abs_alpha - abs_alpha

    # gradients from tf.where can be NaN if the non-selected branch is NaN
    # https://github.com/tensorflow/tensorflow/issues/42889
    # solution is to provide save values for the non-selected branch to never make them become NaNs
    b_tmin = b - tmin
    safe_b_tmin_ones = znp.where(b_tmin > 0, b_tmin, znp.ones_like(b_tmin))
    b_tmax = b - tmax
    safe_b_tmax_ones = znp.where(b_tmax > 0, b_tmax, znp.ones_like(b_tmax))

    if_true_1 = a * abs_sigma * (znp.log(safe_b_tmin_ones) - znp.log(safe_b_tmax_ones))

    if_false_1 = (
        a
        * abs_sigma
        / (1.0 - n)
        * (
            1.0 / znp.power(safe_b_tmin_ones, n - 1.0)
            - 1.0 / znp.power(safe_b_tmax_ones, n - 1.0)
        )
    )

    if_true_3 = tf.where(use_log, if_true_1, if_false_1)

    if_true_2 = a * abs_sigma * (znp.log(safe_b_tmin_ones) - znp.log(n / abs_alpha))
    if_false_2 = (
        a
        * abs_sigma
        / (1.0 - n)
        * (
            1.0 / znp.power(safe_b_tmin_ones, n - 1.0)
            - 1.0 / znp.power(n / abs_alpha, n - 1.0)
        )
    )
    term1 = tf.where(use_log, if_true_2, if_false_2)
    term2 = (
        abs_sigma
        * sqrt_pi_over_two
        * (tf.math.erf(tmax / sqrt2) - tf.math.erf(-abs_alpha / sqrt2))
    )
    if_false_3 = term1 + term2

    if_false_4 = tf.where(tf.less_equal(tmax, -abs_alpha), if_true_3, if_false_3)

    # if_false_4()
    result = tf.where(tf.greater_equal(tmin, -abs_alpha), if_true_4, if_false_4)
    if not result.shape.rank == 0:
        result = tf.gather(result, 0, axis=-1)  # remove last dim, should vanish
    return result


def double_crystalball_mu_integral(limits, params, model):
    mu = params["mu"]
    sigma = params["sigma"]
    alphal = params["alphal"]
    nl = params["nl"]
    alphar = params["alphar"]
    nr = params["nr"]

    lower, upper = limits._rect_limits_tf
    lower = lower[:, 0]
    upper = upper[:, 0]

    return double_crystalball_mu_integral_func(
        mu=mu,
        sigma=sigma,
        alphal=alphal,
        nl=nl,
        alphar=alphar,
        nr=nr,
        lower=lower,
        upper=upper,
    )


@z.function(wraps="zfit_tensor")
def double_crystalball_mu_integral_func(
    mu, sigma, alphal, nl, alphar, nr, lower, upper
):
    # mu_broadcast =
    upper_of_lowerint = znp.minimum(mu, upper)
    integral_left = crystalball_integral_func(
        mu=mu, sigma=sigma, alpha=alphal, n=nl, lower=lower, upper=upper_of_lowerint
    )
    left = tf.where(tf.less(mu, lower), znp.zeros_like(integral_left), integral_left)

    lower_of_upperint = znp.maximum(mu, lower)
    integral_right = crystalball_integral_func(
        mu=mu, sigma=sigma, alpha=-alphar, n=nr, lower=lower_of_upperint, upper=upper
    )
    right = tf.where(
        tf.greater(mu, upper), znp.zeros_like(integral_right), integral_right
    )

    integral = left + right
    return integral


class CrystalBall(BasePDF, SerializableMixin):
    _N_OBS = 1

    def __init__(
        self,
        mu: ztyping.ParamTypeInput,
        sigma: ztyping.ParamTypeInput,
        alpha: ztyping.ParamTypeInput,
        n: ztyping.ParamTypeInput,
        obs: ztyping.ObsTypeInput,
        *,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "CrystalBall",
    ):
        """Crystal Ball shaped PDF. A combination of a Gaussian with a powerlaw tail.

        The function is defined as follows:

        .. math::
            f(x;\\mu, \\sigma, \\alpha, n) =  \\begin{cases} \\exp(- \\frac{(x - \\mu)^2}{2 \\sigma^2}),
            & \\mbox{for}\\frac{x - \\mu}{\\sigma} \\geqslant -\\alpha \\newline
            A \\cdot (B - \\frac{x - \\mu}{\\sigma})^{-n}, & \\mbox{for }\\frac{x - \\mu}{\\sigma}
             < -\\alpha \\end{cases}

        with

        .. math::
            A = \\left(\\frac{n}{\\left| \\alpha \\right|}\\right)^n \\cdot
            \\exp\\left(- \\frac {\\left|\\alpha \\right|^2}{2}\\right)

            B = \\frac{n}{\\left| \\alpha \\right|}  - \\left| \\alpha \\right|

        Args:
            mu: The mean of the gaussian
            sigma: Standard deviation of the gaussian
            alpha: parameter where to switch from a gaussian to the powertail
            n: Exponent of the powertail
            obs: |@doc:pdf.init.obs| Observables of the
               model. This will be used as the default space of the PDF and,
               if not given explicitly, as the normalization range.

               The default space is used for example in the sample method: if no
               sampling limits are given, the default space is used.

               The observables are not equal to the domain as it does not restrict or
               truncate the model outside this range. |@docend:pdf.init.obs|
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:pdf.init.name| Human-readable name
               or label of
               the PDF for better identification.
               Has no programmatical functional purpose as identification. |@docend:pdf.init.name|

        .. _CBShape: https://en.wikipedia.org/wiki/Crystal_Ball_function
        """
        params = {"mu": mu, "sigma": sigma, "alpha": alpha, "n": n}
        super().__init__(
            obs=obs, name=name, params=params, extended=extended, norm=norm
        )

    def _unnormalized_pdf(self, x):
        mu = self.params["mu"].value()
        sigma = self.params["sigma"].value()
        alpha = self.params["alpha"].value()
        n = self.params["n"].value()
        x = x.unstack_x()
        return crystalball_func(x=x, mu=mu, sigma=sigma, alpha=alpha, n=n)


class CrystalBallPDFRepr(BasePDFRepr):
    _implementation = CrystalBall
    hs3_type: Literal["CrystalBall"] = pydantic.Field("CrystalBall", alias="type")
    x: SpaceRepr
    mu: Serializer.types.ParamTypeDiscriminated
    sigma: Serializer.types.ParamTypeDiscriminated
    alpha: Serializer.types.ParamTypeDiscriminated
    n: Serializer.types.ParamTypeDiscriminated


crystalball_integral_limits = Space(
    axes=(0,), limits=(((ANY_LOWER,),), ((ANY_UPPER,),))
)

CrystalBall.register_analytic_integral(
    func=crystalball_integral, limits=crystalball_integral_limits
)


class DoubleCB(BasePDF, SerializableMixin):
    _N_OBS = 1

    def __init__(
        self,
        mu: ztyping.ParamTypeInput,
        sigma: ztyping.ParamTypeInput,
        alphal: ztyping.ParamTypeInput,
        nl: ztyping.ParamTypeInput,
        alphar: ztyping.ParamTypeInput,
        nr: ztyping.ParamTypeInput,
        obs: ztyping.ObsTypeInput,
        *,
        extended: ExtendedInputType = None,
        norm: NormInputType = None,
        name: str = "DoubleCB",
    ):
        """Double-sided Crystal Ball shaped PDF. A combination of two CB using the **mu** (not a frac) on each side.

        The function is defined as follows:

        .. math::
            f(x;\\mu, \\sigma, \\alpha_{L}, n_{L}, \\alpha_{R}, n_{R}) =  \\begin{cases}
            A_{L} \\cdot (B_{L} - \\frac{x - \\mu}{\\sigma})^{-n},
             & \\mbox{for }\\frac{x - \\mu}{\\sigma} < -\\alpha_{L} \\newline
            \\exp(- \\frac{(x - \\mu)^2}{2 \\sigma^2}),
            & -\\alpha_{L} \\leqslant \\mbox{for}\\frac{x - \\mu}{\\sigma} \\leqslant \\alpha_{R} \\newline
            A_{R} \\cdot (B_{R} + \\frac{x - \\mu}{\\sigma})^{-n},
             & \\mbox{for }\\frac{x - \\mu}{\\sigma} > \\alpha_{R}
            \\end{cases}

        with

        .. math::
            A_{L/R} = \\left(\\frac{n_{L/R}}{\\left| \\alpha_{L/R} \\right|}\\right)^n_{L/R} \\cdot
            \\exp\\left(- \\frac {\\left|\\alpha_{L/R} \\right|^2}{2}\\right)

            B_{L/R} = \\frac{n_{L/R}}{\\left| \\alpha_{L/R} \\right|}  - \\left| \\alpha_{L/R} \\right|

        Args:
            mu: The mean of the gaussian
            sigma: Standard deviation of the gaussian
            alphal: parameter where to switch from a gaussian to the powertail on the left
                side
            nl: Exponent of the powertail on the left side
            alphar: parameter where to switch from a gaussian to the powertail on the right
                side
            nr: Exponent of the powertail on the right side
            obs: |@doc:pdf.init.obs| Observables of the
               model. This will be used as the default space of the PDF and,
               if not given explicitly, as the normalization range.

               The default space is used for example in the sample method: if no
               sampling limits are given, the default space is used.

               The observables are not equal to the domain as it does not restrict or
               truncate the model outside this range. |@docend:pdf.init.obs|
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:pdf.init.name| Human-readable name
               or label of
               the PDF for better identification.
               Has no programmatical functional purpose as identification. |@docend:pdf.init.name|
        """
        params = {
            "mu": mu,
            "sigma": sigma,
            "alphal": alphal,
            "nl": nl,
            "alphar": alphar,
            "nr": nr,
        }
        super().__init__(
            obs=obs, name=name, params=params, extended=extended, norm=norm
        )

    def _unnormalized_pdf(self, x):
        mu = self.params["mu"].value()
        sigma = self.params["sigma"].value()
        alphal = self.params["alphal"].value()
        nl = self.params["nl"].value()
        alphar = self.params["alphar"].value()
        nr = self.params["nr"].value()
        x = x.unstack_x()
        return double_crystalball_func(
            x=x, mu=mu, sigma=sigma, alphal=alphal, nl=nl, alphar=alphar, nr=nr
        )


class DoubleCBPDFRepr(BasePDFRepr):
    _implementation = DoubleCB
    hs3_type: Literal["DoubleCB"] = pydantic.Field("DoubleCB", alias="type")
    x: SpaceRepr
    mu: Serializer.types.ParamTypeDiscriminated
    sigma: Serializer.types.ParamTypeDiscriminated
    alphal: Serializer.types.ParamTypeDiscriminated
    nl: Serializer.types.ParamTypeDiscriminated
    alphar: Serializer.types.ParamTypeDiscriminated
    nr: Serializer.types.ParamTypeDiscriminated


DoubleCB.register_analytic_integral(
    func=double_crystalball_mu_integral, limits=crystalball_integral_limits
)



#  Copyright (c) 2023 zfit

from __future__ import annotations

from collections.abc import Iterable

from zfit import z
from zfit.core.interfaces import ZfitPDF
from zfit.util.exception import NotExtendedPDFError
from zfit.z.random import sample_with_replacement, counts_multinomial
from .util.container import convert_to_container

__all__ = ["poisson", "sample_with_replacement", "counts_multinomial"]


def poisson(n=None, pdfs: Iterable[ZfitPDF] = None):
    if n and pdfs:
        raise ValueError("Cannot specify both, `n`, and `pdfs`, at the same time.")

    if pdfs:
        pdfs = convert_to_container(pdfs)
        not_extended = [pdf.is_extended for pdf in pdfs]
        if not_extended:
            raise NotExtendedPDFError(
                f"The following pdfs are not extended but need to be: {not_extended}"
            )
        if len(pdfs) > 1:
            raise ValueError("More than one model (currently) not supported.")

        # single pdf only implementation here
        yield_ = pdfs[0].get_yield()
    else:
        yield_ = n

    poisson_term = z.random.poisson(lam=yield_)
    return poisson_term



#  Copyright (c) 2023 zfit
import pytest


@pytest.mark.parametrize("n", [500, 10000, 200_000], ids=lambda x: f"samplesize={x:,}")
@pytest.mark.parametrize(
    "floatall", [True, False], ids=["floatAll", "floatSigBkgLambdaOnly"]
)
@pytest.mark.parametrize(
    "use_sampler", [True, False], ids=["useSampler", "onetimeSample"]
)
@pytest.mark.parametrize("nbins", [27, 90, 311], ids=lambda x: f"nbins={x:,}")
@pytest.mark.parametrize(
    "use_wrapper",
    [True, False],
    ids=lambda x: f"useWrapper={x}",
)
@pytest.mark.flaky(reruns=1)  # in case a fit fails
def test_sig_bkg_fit(n, floatall, use_sampler, nbins, use_wrapper, request):
    import mplhep
    import matplotlib.pyplot as plt
    import numpy as np

    import zfit

    longtest = request.config.getoption("--longtests")
    if not longtest:
        if n > 1000 or floatall or use_sampler:
            pytest.skip("skipping long test")

    if nbins < 90 or n < 1000:
        zfit.settings.set_seed(42)

    # create space
    obs_binned = zfit.Space(
        "x", binning=zfit.binned.RegularBinning(nbins, -10, 10, name="x")
    )
    obs = obs_binned.with_binning(None)

    # parameters
    mu = zfit.Parameter("mu", 1.0, -4, 6, floating=floatall)
    sigma = zfit.Parameter("sigma", 1.0, 0.1, 10, floating=floatall)

    lambd = zfit.Parameter("lambda", -0.06, -1, -0.01)

    # model building, pdf creation
    gauss = zfit.pdf.Gauss(mu=mu, sigma=sigma, obs=obs)
    exponential = zfit.pdf.Exponential(lambd, obs=obs)

    n_bkg_init = 20000
    n_sig_init = 2000
    n_bkg = zfit.Parameter("n_bkg", n_bkg_init)
    n_sig = zfit.Parameter("n_sig", n_sig_init)
    gauss_extended = gauss.create_extended(n_sig)
    exp_extended = exponential.create_extended(n_bkg)
    model_unbinned = zfit.pdf.SumPDF([gauss_extended, exp_extended])

    # make binned
    if use_wrapper:
        model = zfit.pdf.BinnedFromUnbinnedPDF(model_unbinned, space=obs_binned)
    else:
        model = model_unbinned.to_binned(obs_binned)

    # data
    n_sample = n
    if use_sampler:
        data = model.create_sampler(n=n_sample)
        data.resample(n=n_sample)
        data_unbinned = model_unbinned.create_sampler(n=n_sample)
        data_unbinned.resample(n=n_sample)
    else:
        data = model.sample(n=n_sample)
        data_unbinned = model_unbinned.sample(n=n_sample)

    plot_scaling = n_sample / nbins * obs.area()

    x = np.linspace(-10, 10, 1000)

    plot_folder = (
        f'NLL_profile/nbins{nbins} n{n} floatall_{floatall}{" sampler" if use_sampler else ""}'
        f'{" wrapper" if use_wrapper else ""}'
    ).replace(" ", "_")

    def plot_pdf(title):
        plt.figure()
        plt.title(title)
        y = model.pdf(x).numpy()
        y_gauss = (gauss.pdf(x) * model_unbinned.params["frac_0"]).numpy()
        y_exp = (exponential.pdf(x) * model_unbinned.params["frac_1"]).numpy()
        plt.plot(x, y * plot_scaling, label="Sum - Model")
        plt.plot(x, y_gauss * plot_scaling, label="Gauss - Signal")
        plt.plot(x, y_exp * plot_scaling, label="Exp - Background")
        mplhep.histplot(data.to_hist(), yerr=True, color="black", histtype="errorbar")
        plt.ylabel("Counts")
        plt.xlabel("obs: $B_{mass}$")
        plt.legend()
        pytest.zfit_savefig(folder=plot_folder)

    # create NLL
    nll = zfit.loss.ExtendedBinnedNLL(model=model, data=data)
    nll_unbinned = zfit.loss.ExtendedUnbinnedNLL(
        model=model_unbinned, data=data_unbinned
    )

    # create a minimizer
    minimizer = zfit.minimize.Minuit()
    results = []
    rel = 0.03 * n**0.5 / 17  # 17 is 300 ** 0.5
    if floatall:
        rel *= 3  # higher tolerance if we float all
    rel *= 300 / nbins
    for loss in [nll, nll_unbinned]:
        if floatall:
            mu.set_value(0.5)
            sigma.set_value(1.2)
        lambd.set_value(-0.05)
        if not use_sampler:
            plot_pdf(
                f"before fit - {loss.name} n:{n_sample:,} {'floatAll' if floatall else 'fixedSigShape'}"
            )

        result = minimizer.minimize(nll)
        results.append(result)
        # do the error calculations, here with hesse, than with minos
        _ = result.hesse()
        (
            param_errors,
            _,
        ) = result.errors()  # this returns a new FitResult if a new minimum was found
        if not use_sampler:
            plot_pdf(
                f"after fit - {loss.name} n:{n_sample:,} {'floatAll' if floatall else 'fixedSigShape'}"
            )

    mu.floating = True
    minimizer.minimize(nll_unbinned)
    mu.floating = False
    param_vals = np.linspace(mu.value() - 0.3, mu.value() + 0.3)
    nlls = []
    nlls_binned = []
    for val in param_vals:
        mu.set_value(val)
        try:
            minimizer.minimize(nll_unbinned)
        except Exception:
            continue

        nlls.append(nll_unbinned.value())
        try:
            minimizer.minimize(nll)
        except Exception:
            nlls.pop(-1)
            continue
        nlls_binned.append(nll.value())

    plt.figure()
    nlls = np.array(nlls) - np.min(nlls)
    nlls_binned = np.array(nlls_binned) - np.min(nlls_binned)
    plt.title(
        f'NLL profile: nbins:{nbins} n:{n} floatall:{floatall} {"sampler" if use_sampler else ""} {"wrapper" if use_wrapper else ""}'
    )
    plt.plot(param_vals, nlls, label="Unbinned")
    plt.plot(param_vals, nlls_binned, label="Binned")
    plt.legend()
    pytest.zfit_savefig(folder=plot_folder)

    assert (
        pytest.approx(results[0].params["lambda"]["value"], rel=rel)
        == results[1].params["lambda"]["value"]
    )
    assert (
        pytest.approx(results[0].params["lambda"]["hesse"], rel=rel)
        == results[1].params["lambda"]["hesse"]
    )
    assert (
        pytest.approx(results[0].params["lambda"]["errors"]["lower"], rel=rel)
        == results[1].params["lambda"]["errors"]["lower"]
    )
    assert (
        pytest.approx(results[0].params["lambda"]["errors"]["upper"], rel=rel)
        == results[1].params["lambda"]["errors"]["upper"]
    )
    assert (
        pytest.approx(results[0].params["n_bkg"]["value"], rel=rel)
        == results[1].params["n_bkg"]["value"]
    )
    assert (
        pytest.approx(results[0].params["n_bkg"]["hesse"], rel=rel)
        == results[1].params["n_bkg"]["hesse"]
    )
    assert (
        pytest.approx(results[0].params["n_bkg"]["errors"]["lower"], rel=rel)
        == results[1].params["n_bkg"]["errors"]["lower"]
    )
    assert (
        pytest.approx(results[0].params["n_bkg"]["errors"]["upper"], rel=rel)
        == results[1].params["n_bkg"]["errors"]["upper"]
    )
    assert (
        pytest.approx(results[0].params["n_sig"]["value"], rel=rel)
        == results[1].params["n_sig"]["value"]
    )
    assert (
        pytest.approx(results[0].params["n_sig"]["hesse"], rel=rel)
        == results[1].params["n_sig"]["hesse"]
    )
    assert (
        pytest.approx(results[0].params["n_sig"]["errors"]["lower"], rel=rel)
        == results[1].params["n_sig"]["errors"]["lower"]
    )
    assert (
        pytest.approx(results[0].params["n_sig"]["errors"]["upper"], rel=rel)
        == results[1].params["n_sig"]["errors"]["upper"]
    )

    if floatall:
        assert (
            pytest.approx(results[0].params["mu"]["value"], rel=rel)
            == results[1].params["mu"]["value"]
        )
        assert (
            pytest.approx(results[0].params["mu"]["hesse"], rel=rel)
            == results[1].params["mu"]["hesse"]
        )
        assert (
            pytest.approx(results[0].params["mu"]["errors"]["lower"], rel=rel)
            == results[1].params["mu"]["errors"]["lower"]
        )
        assert (
            pytest.approx(results[0].params["mu"]["errors"]["upper"], rel=rel)
            == results[1].params["mu"]["errors"]["upper"]
        )
        assert (
            pytest.approx(results[0].params["sigma"]["value"], rel=rel)
            == results[1].params["sigma"]["value"]
        )
        assert (
            pytest.approx(results[0].params["sigma"]["hesse"], rel=rel)
            == results[1].params["sigma"]["hesse"]
        )
        assert (
            pytest.approx(results[0].params["sigma"]["errors"]["lower"], rel=rel)
            == results[1].params["sigma"]["errors"]["lower"]
        )
        assert (
            pytest.approx(results[0].params["sigma"]["errors"]["upper"], rel=rel)
            == results[1].params["sigma"]["errors"]["upper"]
        )


def test_nbins(request):
    #  Copyright (c) 2022 zfit

    import matplotlib.pyplot as plt
    import mplhep
    import numpy as np
    import zfit.z.numpy as znp

    import zfit

    longtest = request.config.getoption("--longtests")

    # create space
    obs = zfit.Space("x", limits=(0, 10))

    # parameters
    init_vals = [3.1, 1.0, -0.06]
    mu = zfit.Parameter("mu", init_vals[0], -4, 10, floating=False)
    sigma = zfit.Parameter("sigma", init_vals[1], 0.1, 10, floating=False)
    lambd = zfit.Parameter("lambda", init_vals[2], -1, -0.01)

    # model building, pdf creation
    gauss = zfit.pdf.Gauss(mu=mu, sigma=sigma, obs=obs)
    exponential = zfit.pdf.Exponential(lambd, obs=obs)

    n_bkg = zfit.Parameter("n_bkg", 300)
    n_sig = zfit.Parameter("n_sig", 20)
    gauss_extended = gauss.create_extended(n_sig)
    exp_extended = exponential.create_extended(n_bkg)
    model_unbinned = zfit.pdf.SumPDF([gauss_extended, exp_extended])
    sig_data = model_unbinned.sample(275)
    bkg_data = model_unbinned.sample(25)
    data = zfit.Data.from_tensor(
        obs, tensor=znp.concatenate([sig_data.value(), bkg_data.value()], axis=0)
    )
    # make binned
    plot_folder = "nbins_accuracy"
    binnings = [5, 50, 1000]
    if longtest:
        binnings += [7, 10, 15, 20, 30, 102, 203, 341, 500]
        binnings = sorted(binnings)
    minimizer = zfit.minimize.Minuit()
    n_bkg_vals = []
    n_bkg_vals_unbinned = []

    for nbins in binnings:
        n_sig.floating = True
        obs_binned = obs.with_binning(nbins)
        model_binned = zfit.pdf.BinnedFromUnbinnedPDF(model_unbinned, space=obs_binned)
        data_binned = data.to_binned(obs_binned)
        # fit
        loss = zfit.loss.ExtendedUnbinnedNLL(model_unbinned, data)
        result_unbinned = minimizer.minimize(loss)
        n_bkg_vals_unbinned.append(result_unbinned.params["n_bkg"]["value"])
        loss_binned = zfit.loss.ExtendedBinnedNLL(model_binned, data_binned)
        result = minimizer.minimize(loss_binned)
        n_bkg_vals.append(result.params["n_bkg"]["value"])
        # plot
        plt.figure()
        plt.title(f"{nbins} bins full fit")
        mplhep.histplot(data_binned, alpha=0.5, yerr=False, label="data")
        mplhep.histplot(model_binned.to_hist(), label="model")
        plt.legend()
        pytest.zfit_savefig(folder=plot_folder)

        plt.figure()
        plt.title(f"{nbins} bins binned vs unbinned curve")
        x = np.linspace(0, 10, nbins)
        scaled_density = model_binned.to_hist().density() * model_binned.get_yield()
        plt.plot(x, scaled_density, "x", label="binned")
        plt.plot(x, model_unbinned.ext_pdf(x), label="unbinned")
        plt.xlabel("x")
        plt.ylabel("density")
        plt.legend()
        pytest.zfit_savefig(folder=plot_folder)
        # plt.show()

        if not result.valid:
            continue

    plt.figure()
    plt.title("n_bkg vs nbins")
    plt.semilogx(binnings, n_bkg_vals, "x", label="binned")
    mean = np.mean(n_bkg_vals_unbinned)
    std = np.std(n_bkg_vals_unbinned)
    plt.semilogx([np.min(binnings), np.max(binnings)], "b", [mean] * 2, label="true")
    plt.semilogx(
        [np.min(binnings), np.max(binnings)], "r", [mean] * 2, label="unbinned"
    )
    plt.semilogx(
        [np.min(binnings), np.max(binnings)],
        [mean + std] * 2,
        "r",
        alpha=0.5,
        label="unbinned + std",
    )
    plt.semilogx(
        [np.min(binnings), np.max(binnings)],
        [mean - std] * 2,
        "r",
        alpha=0.5,
        label="unbinned - std",
    )
    plt.xlabel("$N_{bins}$")
    plt.ylabel("$N_{bkg}$")
    plt.legend()
    pytest.zfit_savefig(folder=plot_folder)
    # plt.show()
    # 4 sigma away, factor of two because binned is less precise
    assert pytest.approx(mean, abs=std * 4 * 2) == result.params["n_bkg"]["value"]



#  Copyright (c) 2023 zfit
import json

import asdf
import numpy as np
import pytest


def test_dumpload_hs3_pdf(request):
    import zfit

    mu = zfit.Parameter("mu", 1.2, -4.0, 5.0)
    sigma3 = zfit.Parameter("sigma3", 0.1, 0.0, 1.0)
    sigma2 = zfit.Parameter("sigma2", 42, 0.0, 100.0)
    sigma1_free = zfit.Parameter("sigma1_free", 422, 0.0, 1002.0)
    sigma1 = zfit.ComposedParameter("sigma1", lambda x: x + 1, params=sigma1_free)
    mu3 = zfit.Parameter("mu3", 2, -2, 3, step_size=0.1)
    frac1 = zfit.param.ConstantParameter("frac1", 0.2)
    frac2 = zfit.Parameter("frac2", 0.3, 0.0, 1.0)

    obs = zfit.Space("obs1", limits=(-4, 5))
    gauss1 = zfit.pdf.Gauss(mu=mu, sigma=sigma1, obs=obs)
    gauss2 = zfit.pdf.Gauss(mu=mu, sigma=sigma2, obs=obs)
    gauss3 = zfit.pdf.Gauss(mu=mu3, sigma=sigma3, obs=obs)
    model = zfit.pdf.SumPDF([gauss1, gauss2, gauss3], fracs=[frac1, frac2])

    hs3model = zfit.hs3.dumps(model)
    hs3model["metadata"]["serializer"]["version"] = "ZFIT_ARBITRARY_VALUE"
    hs3model_true = pytest.helpers.get_truth("hs3", "sum3gauss.json", request, hs3model)
    hs3model_cleaned, hs3model_true = pytest.helpers.cleanup_hs3(
        hs3model, hs3model_true
    )
    assert hs3model_cleaned == hs3model_true

    model_loaded = zfit.hs3.loads(hs3model)
    assert model_loaded is not model
    hs3model = zfit.hs3.dumps(model_loaded["distributions"].values())
    hs3model_cleaned, hs3model_true = pytest.helpers.cleanup_hs3(
        hs3model, hs3model_true
    )
    assert hs3model_cleaned == hs3model_true
    # second time, no loop to better check if failure in unittest
    model_loaded = zfit.hs3.loads(hs3model)
    assert model_loaded is not model
    model_load = list(model_loaded["distributions"].values())
    hs3model = zfit.hs3.dumps(model_load)

    x = np.linspace(-4, 5, 100)
    np.testing.assert_allclose(model.pdf(x), model_load[0].pdf(x))

    hs3model_cleaned, hs3model_true = pytest.helpers.cleanup_hs3(
        hs3model, hs3model_true
    )
    assert hs3model_cleaned == hs3model_true

    hs3model_false_truth = hs3model_true.copy()
    hs3model_false_truth["variables"]["mu3"] = mu


def test_dumpload_hs3_loss(request):
    import zfit

    mu = zfit.Parameter("mu", 1.2, -4.0, 5.0)
    sigma3 = zfit.Parameter("sigma3", 0.6, 0.0, 1.0)
    sigma2 = zfit.Parameter("sigma2", 0.4, 0.0, 100.0)
    sigma4 = zfit.Parameter("sigma4", 0.5, 0.0, 1.0)
    sigma1_free = zfit.Parameter("sigma1_free", 422, 0.0, 1002.0)
    sigma1 = zfit.ComposedParameter("sigma1", lambda x: x + 1, params=sigma1_free)
    mu3 = zfit.Parameter("mu3", 2, -2, 3, step_size=0.1)
    mu4 = zfit.Parameter("mu4", 2, -2, 3, step_size=0.1)
    frac1 = zfit.param.ConstantParameter("frac1", 0.2)
    frac2 = zfit.Parameter("frac2", 0.3, 0.0, 1.0)
    frac3 = zfit.Parameter("frac3", 0.4, 0.0, 1.0)

    obs = zfit.Space("obs1", limits=(-4, 5))
    gauss1 = zfit.pdf.Gauss(mu=mu, sigma=sigma1, obs=obs)
    gauss2 = zfit.pdf.Gauss(mu=mu, sigma=sigma2, obs=obs)
    gauss3 = zfit.pdf.Gauss(mu=mu3, sigma=sigma3, obs=obs)
    gauss4 = zfit.pdf.Gauss(mu=mu4, sigma=sigma4, obs=obs)
    model = zfit.pdf.SumPDF([gauss1, gauss2, gauss3], fracs=[frac1, frac2])
    model2 = zfit.pdf.SumPDF([model, gauss4], fracs=frac3)

    data1 = zfit.data.Data.from_numpy(obs=obs, array=np.random.normal(1.2, 1, 17))
    data2 = zfit.data.Data.from_numpy(obs=obs, array=np.random.normal(1.31, 1, 9))
    data3 = zfit.data.Data.from_numpy(obs=obs, array=np.random.normal(1.4, 2, 14))
    constraint = zfit.constraint.GaussianConstraint(
        params=mu, observation=1.2, uncertainty=0.14
    )
    loss = zfit.loss.UnbinnedNLL(
        model=[model, gauss2, model2],
        data=[data1, data2, data3],
        constraints=constraint,
    )
    original_lossval = loss.value(full=False)
    original_loss = loss

    hs3model = zfit.hs3.dumps(loss)
    print(hs3model)
    hs3model["metadata"]["serializer"]["version"] = "ZFIT_ARBITRARY_VALUE"
    hs3model_asdf = asdf.AsdfFile(hs3model)
    hs3model_true = pytest.helpers.get_truth(
        "hs3", "UnbinnedNLL_simple.asdf", request, hs3model_asdf
    ).tree
    hs3model_cleaned, hs3model_true = pytest.helpers.cleanup_hs3(
        hs3model, hs3model_true
    )
    hs3model_cleaned.pop("asdf_library", None)
    hs3model_true.pop("asdf_library", None)
    hs3model_cleaned.pop("history", None)
    hs3model_true.pop("history", None)

    for key, val in hs3model_cleaned.items():
        if isinstance(val, np.ndarray):
            np.testing.assert_equal(val, hs3model_true[key])
        else:
            for k, v in val.items():
                if isinstance(v, np.ndarray):
                    np.testing.assert_equal(v, val[k])
                else:
                    try:
                        is_equal = v == val[k]
                    except ValueError:
                        pass
                    else:
                        if not is_equal:
                            assert v == val[k]  # make sure the diff is shown by pytest
    model_loaded = zfit.hs3.loads(hs3model)
    loss = list(model_loaded["loss"].values())[0]
    assert pytest.approx(loss.value(full=False), rel=1e-3) == original_lossval
    for k, v in hs3model_cleaned.items():
        if isinstance(v, np.ndarray):
            np.testing.assert_equal(v, hs3model_true[k])
        else:
            for k, v in val.items():
                if isinstance(v, np.ndarray):
                    np.testing.assert_equal(v, val[k])
                else:
                    try:
                        is_equal = v == val[k]
                    except ValueError:
                        pass
                    else:
                        if not is_equal:
                            assert v == val[k]
    # np.testing.assert_equal(hs3model_cleaned, hs3model_true)

    # assert hs3model_cleaned == hs3model_true

    model_loaded = zfit.hs3.loads(hs3model)
    assert model_loaded is not model
    models_load = list(model_loaded["distributions"].values())
    loss_loaded = list(model_loaded["loss"].values())[0]
    x = np.linspace(-4, 5, 100)
    for modelload, modeltrue in zip(loss_loaded.model, original_loss.model):
        np.testing.assert_allclose(modelload.pdf(x), modeltrue.pdf(x))
    hs3model = zfit.hs3.dumps(models_load)
    hs3model_cleaned, hs3model_true = pytest.helpers.cleanup_hs3(
        hs3model, hs3model_true
    )
    for k, v in hs3model_cleaned.items():
        if isinstance(v, np.ndarray):
            np.testing.assert_equal(v, hs3model_true[k])
        else:
            for k, v in val.items():
                if isinstance(v, np.ndarray):
                    np.testing.assert_equal(v, val[k])
                else:
                    try:
                        is_equal = v == val[k]
                    except ValueError:
                        pass
                    else:
                        if not is_equal:
                            assert v == val[k]
    # second time, no loop to better check if failure in unittest
    model_loaded = zfit.hs3.loads(hs3model)
    assert model_loaded is not model
    hs3model = zfit.hs3.dumps(model_loaded["distributions"].values())

    hs3model_cleaned, hs3model_true = pytest.helpers.cleanup_hs3(
        hs3model, hs3model_true
    )
    for k, v in hs3model_cleaned.items():
        if isinstance(v, np.ndarray):
            np.testing.assert_equal(v, hs3model_true[k])
        else:
            for k, v in val.items():
                if isinstance(v, np.ndarray):
                    np.testing.assert_equal(v, val[k])
                else:
                    try:
                        is_equal = v == val[k]
                    except ValueError:
                        pass
                    else:
                        if not is_equal:
                            assert v == val[k]



#  Copyright (c) 2023 zfit

from __future__ import annotations

from collections.abc import Iterable

from .basefunctor import FunctorMixin, _preprocess_init_sum
from .. import z
from ..core.binnedpdf import BaseBinnedPDFV1
from ..core.interfaces import ZfitPDF
from ..core.space import supports
from ..util import ztyping
from ..util.container import convert_to_container
from ..util.deprecation import deprecated_norm_range
from ..util.exception import NormNotImplemented
from ..z import numpy as znp


class BaseBinnedFunctorPDF(FunctorMixin, BaseBinnedPDFV1):
    """Base class for binned functors."""

    def __init__(self, models, obs, **kwargs):
        super().__init__(models, obs, **kwargs)
        self.pdfs = self.models


class BinnedSumPDF(BaseBinnedFunctorPDF):
    def __init__(
        self,
        pdfs: Iterable[ZfitPDF],
        fracs: ztyping.ParamTypeInput | None = None,
        obs: ztyping.ObsTypeInput = None,
        name: str = "BinnedSumPDF",
    ):
        self._fracs = None

        pdfs = convert_to_container(pdfs)
        self.pdfs = pdfs

        (
            all_extended,
            fracs_cleaned,
            param_fracs,
            params,
            sum_yields,
            frac_param_created,
        ) = _preprocess_init_sum(fracs, obs, pdfs)
        del frac_param_created  # currently actually not used

        self._fracs = param_fracs
        self._original_fracs = fracs_cleaned

        extended = sum_yields if all_extended else None
        super().__init__(
            models=pdfs, obs=obs, params=params, name=name, extended=extended
        )

    # def _unnormalized_pdf(self, x):
    #     models = self.models
    #     prob = tf.reduce_sum([model._unnormalized_pdf(x) for model in models], axis=0)
    #     return prob

    @supports(norm=True)
    def _pdf(self, x, norm):
        equal_norm_ranges = len(set([pdf.norm for pdf in self.pdfs] + [norm])) == 1
        if norm and not equal_norm_ranges:
            raise NormNotImplemented
        pdfs = self.pdfs
        fracs = self.params.values()
        probs = []
        for pdf, frac in zip(pdfs, fracs):
            probs.append(pdf.pdf(x) * frac)
        prob = znp.sum(probs, axis=0)
        return z.convert_to_tensor(prob)

    @deprecated_norm_range
    def _ext_pdf(self, x, norm, *, norm_range=None):
        equal_norm_ranges = len(set([pdf.norm for pdf in self.pdfs] + [norm])) == 1
        if norm and not equal_norm_ranges:
            raise NormNotImplemented
        prob = znp.sum([model.ext_pdf(x) for model in self.models], axis=0)
        return z.convert_to_tensor(prob)

    def _counts(self, x, norm=None):
        equal_norm_ranges = len(set([pdf.norm for pdf in self.pdfs] + [norm])) == 1
        if norm and not equal_norm_ranges:
            raise NormNotImplemented
        prob = znp.sum([model.counts(x) for model in self.models], axis=0)
        return prob

    def _rel_counts(self, x, norm=None):
        equal_norm_ranges = len(set([pdf.norm for pdf in self.pdfs] + [norm])) == 1
        if norm and not equal_norm_ranges:
            raise NormNotImplemented
        fracs = self.params.values()
        prob = znp.sum(
            [model.rel_counts(x) * frac for model, frac in zip(self.models, fracs)],
            axis=0,
        )
        return prob



#  Copyright (c) 2024 zfit

from __future__ import annotations

import contextlib
from collections.abc import Callable
from functools import partial

import numpy as np
import tensorflow as tf
import texttable as tt

from .strategy import ZfitStrategy
from .. import z
from ..core.interfaces import ZfitLoss
from ..core.parameter import assign_values_jit, assign_values
from ..util import ztyping
from ..util.container import convert_to_container
from ..util.exception import DerivativeCalculationError, MaximumIterationReached
from ..z import numpy as znp


def assign_values_func(params, values):
    # TODO(WrappedVariable): this is needed if we want to use wrapped Variables
    # params = z.math._extract_tfparams(params)
    return assign_values_jit(params, znp.asarray(values))


def check_derivative_none_raise(values, params) -> None:
    """Check if values contains ``None`` and raise a ``DerivativeCalculationError`` if so.

    Args:
        values: Values to check. Can be gradient or hessian
        params: Parameter that correspond to the values.
    """
    if None in values:
        none_params = [p for p, grad in zip(params, values) if grad is None]
        raise DerivativeCalculationError(
            f"The derivative of the following parameters is None: {none_params}."
            f" This is usually caused by either the function not depending on the"
            f" parameter (or not in a differentiable way) or by using pure Python"
            f" code instead of TensorFlow."
        )


class LossEval:
    def __init__(
        self,
        loss: ZfitLoss,
        params: ztyping.ParamTypeInput,
        strategy: ZfitStrategy,
        do_print: bool,
        maxiter: int,
        grad_fn: Callable | None = None,
        hesse_fn: Callable | None = None,
        numpy_converter: Callable | None = None,
        full: bool | None = None,
    ):
        r"""Convenience wrapper for the evaluation of a loss with given parameters and strategy.

        The methods ``value``, ``gradient`` etc will raise a ``MaximumIterationReached`` error in case the maximum iterations
        is reached.

        Args:
            loss: Loss that will be used to be evaluated.
            params: Parameters that the gradient and hessian will be derived to.
            strategy: A strategy to deal with NaNs and more.
            do_print: If the values should be printed nicely
            maxiter: Maximum number of evaluations of the ``value``, 'gradient`` or ``hessian``.
            grad_fn: Function that returns the gradient.
            hesse_fn: Function that returns the hessian matrix.
            numpy_converter: Converter to a numpy-like format. For example, ``np.array`` will create
                a *writable* numpy array, whereas ``np.asarray`` will create a *read-only* numpy array.
                Useful if the return values should be numpy arrays (and not "numpy-like" objects). This is needed
                if the function expects something *writeable* like a numpy array as other (JAX, TensorFlow) arrays
                are not writeable. If None, no conversion is done and a "nmupy-like" object is returned.
            full: If True, the full loss will be calculated. Default is False. For the minimization, the full loss is usually not needed.
        """
        super().__init__()
        full = False if full is None else full
        self.maxiter = maxiter
        self._ignoring_maxiter = False
        with self.ignore_maxiter():  # otherwise when trying to set, it gets all of them, which fails as not there
            self.nfunc_eval = 0
            self.ngrad_eval = 0
            self.nhess_eval = 0

        self.loss = loss
        if hesse_fn is None:
            hesse_fn = loss.hessian

        self.hesse_fn = hesse_fn
        if grad_fn is not None:

            def value_gradients_fn(params):
                return loss.value(full=self.full), grad_fn(params)

        else:
            value_gradients_fn = partial(self.loss.value_gradient, full=False)
            grad_fn = self.loss.gradient
        self.gradients_fn = grad_fn
        self.value_gradients_fn = value_gradients_fn
        self.loss = loss
        self.last_value = None
        self.last_gradient = None
        self.last_hessian = None
        self.nan_counter = 0
        self.params = convert_to_container(params)
        self.strategy = strategy
        self.do_print = do_print
        self.full = full
        self.numpy_converter = False if numpy_converter is None else numpy_converter

    @property
    def niter(self):
        return max([self.nfunc_eval, self.ngrad_eval, self.nhess_eval])

    @property
    def maxiter_reached(self):
        return not self.ignoring_maxiter and self.niter > self.maxiter

    def _check_maxiter_reached(self):
        if self.maxiter_reached:
            raise MaximumIterationReached

    @contextlib.contextmanager
    def ignore_maxiter(self):
        """Return temporary the maximum number of iterations and won't raise an error."""
        old = self._ignoring_maxiter
        self._ignoring_maxiter = True
        yield
        self._ignoring_maxiter = old

    @property
    def ignoring_maxiter(self):
        return self._ignoring_maxiter or self.maxiter is None

    @property
    def nfunc_eval(self):
        return self._nfunc_eval

    @nfunc_eval.setter
    def nfunc_eval(self, value):
        self._nfunc_eval = value
        self._check_maxiter_reached()

    @property
    def ngrad_eval(self):
        return self._ngrad_eval

    @ngrad_eval.setter
    def ngrad_eval(self, value):
        self._ngrad_eval = value
        self._check_maxiter_reached()

    @property
    def nhess_eval(self):
        return self._nhess_eval

    @nhess_eval.setter
    def nhess_eval(self, value):
        self._nhess_eval = value
        self._check_maxiter_reached()

    def value_gradient(self, values: np.ndarray) -> tuple[np.float64, np.ndarray]:
        """Calculate the value and gradient like :py:meth:`~ZfitLoss.value_gradients`.

        Args:
            values: parameter values to calculate the value and gradient at.

        Returns:
            Tuple[numpy.float64, numpy.ndarray]:

        Raises:
            MaximumIterationReached: If the maximum number of iterations (including tolerance) was reached.
        """
        if not self._ignoring_maxiter:
            self.nfunc_eval += 1
            self.ngrad_eval += 1

        params = self.params
        assign_values_func(params, values=values)
        is_nan = False

        try:
            loss_value, gradient = self.value_gradients_fn(params=params)
            loss_value, gradient, _ = self.strategy.callback(
                value=loss_value,
                gradient=gradient,
                hessian=None,
                params=params,
                loss=self.loss,
            )
        except Exception as error:
            loss_value = "invalid, error occured"
            gradient = ["invalid"] * len(params)
            if isinstance(error, tf.errors.InvalidArgumentError):
                is_nan = True
            else:
                raise

        finally:
            if self.do_print:
                try:
                    print_gradient(params, values, gradient=gradient, loss=loss_value)
                except:
                    print("Cannot print loss value or gradient values.")

        check_derivative_none_raise(gradient, params)
        is_nan = is_nan or any(znp.isnan(gradient)) or znp.isnan(loss_value)
        if is_nan:
            self.nan_counter += 1
            info_values = {
                "loss": loss_value,
                "grad": gradient,
                "old_loss": self.last_value,
                "old_grad": self.last_gradient,
                "nan_counter": self.nan_counter,
            }

            loss_value, gradient = self.strategy.minimize_nan(
                loss=self.loss, params=params, values=info_values
            )
        else:
            self.nan_counter = 0
            self.last_value = loss_value
            self.last_gradient = gradient
        if self.numpy_converter:
            loss_value = float(loss_value)
            gradient = self.numpy_converter(gradient)
        return loss_value, gradient

    def value(self, values: np.ndarray) -> np.float64:
        """Calculate the value like :py:meth:`~ZfitLoss.value`.

        Args:
            values: parameter values to calculate the value at.

        Returns:
            Calculated loss value

        Raises:
            MaximumIterationReached: If the maximum number of iterations (including tolerance) was reached.
        """
        if not self._ignoring_maxiter:
            self.nfunc_eval += 1
        params = self.params
        assign_values_func(params, values=values)
        is_nan = False

        try:
            loss_value = self.loss.value(full=self.full)
            loss_value, _, _ = self.strategy.callback(
                value=loss_value,
                gradient=None,
                hessian=None,
                params=params,
                loss=self.loss,
            )
        except Exception as error:
            loss_value = "invalid, error occured"
            if isinstance(error, tf.errors.InvalidArgumentError):
                is_nan = True
            else:
                raise

        finally:
            if self.do_print:
                try:
                    print_params(params, values, loss=loss_value)
                except:
                    print("Cannot print loss value or gradient values.")

        is_nan = is_nan or np.isnan(loss_value).any()
        if is_nan:
            self.nan_counter += 1
            info_values = {
                "loss": loss_value,
                "old_loss": self.last_value,
                "nan_counter": self.nan_counter,
            }

            loss_value, _ = self.strategy.minimize_nan(
                loss=self.loss, params=params, values=info_values
            )
        else:
            self.nan_counter = 0
            self.last_value = loss_value
        if self.numpy_converter:
            loss_value = float(loss_value)
        return loss_value

    def gradient(self, values: np.ndarray) -> np.ndarray:
        """Calculate the gradient like :py:meth:`~ZfitLoss.gradient`.

        Args:
            values: parameter values to calculate the value and gradient at.

        Returns:
            Tuple[numpy.float64, numpy.ndarray]:

        Raises:
            MaximumIterationReached: If the maximum number of iterations (including tolerance) was reached.
        """
        if not self._ignoring_maxiter:
            self.ngrad_eval += 1
        params = self.params
        assign_values_func(params, values=values)
        is_nan = False

        try:
            gradient = self.gradients_fn(params=params)
            _, gradient, _ = self.strategy.callback(
                value=None,
                gradient=gradient,
                hessian=None,
                params=params,
                loss=self.loss,
            )
        except Exception as error:
            gradient = ["invalid"] * len(params)
            if isinstance(error, tf.errors.InvalidArgumentError):
                is_nan = True
            else:
                raise

        finally:
            if self.do_print:
                try:
                    print_gradient(params, values, gradient=gradient, loss=None)
                except:
                    print("Cannot print loss value or gradient values.")

        check_derivative_none_raise(gradient, params)

        is_nan = is_nan or znp.any(znp.isnan(gradient))
        if is_nan:
            self.nan_counter += 1
            info_values = {
                "loss": None,
                "grad": gradient,
                "old_loss": self.last_value,
                "old_grad": self.last_gradient,
                "nan_counter": self.nan_counter,
            }

            _, gradient = self.strategy.minimize_nan(
                loss=self.loss, params=params, values=info_values
            )
        else:
            self.nan_counter = 0
            self.last_gradient = gradient
        if self.numpy_converter:
            gradient = self.numpy_converter(gradient)
        return gradient

    def hessian(self, values) -> znp.array:
        """Calculate the hessian like :py:meth:`~ZfitLoss.hessian`.

        Args:
            values: parameter values to calculate the hessian at.

        Returns:
            Hessian matrix

        Raises:
            MaximumIterationReached: If the maximum number of iterations (including tolerance) was reached.
        """
        if not self._ignoring_maxiter:
            self.nhess_eval += 1
        params = self.params
        assign_values_func(params, values=values)
        is_nan = False

        try:
            hessian = self.hesse_fn(params=params)
            _, _, hessian = self.strategy.callback(
                value=None,
                gradient=None,
                hessian=hessian,
                params=params,
                loss=self.loss,
            )
        except Exception as error:
            hessian = ["invalid"] * len(params)
            if isinstance(error, tf.errors.InvalidArgumentError):
                is_nan = True
            else:
                raise

        finally:
            if self.do_print:
                try:
                    print_params(params, values, loss=None)
                except:
                    print("Cannot print loss value or gradient values.")
        check_derivative_none_raise(hessian, params)

        is_nan = is_nan or znp.any(znp.isnan(hessian))
        if is_nan:
            self.nan_counter += 1
            info_values = {
                "loss": None,
                "old_loss": self.last_value,
                "old_grad": self.last_gradient,
                "nan_counter": self.nan_counter,
            }

            _, _ = self.strategy.minimize_nan(
                loss=self.loss, params=params, values=info_values
            )
        else:
            self.nan_counter = 0
            self.last_hessian = hessian
        if self.numpy_converter:
            hessian = self.numpy_converter(hessian)
        return hessian


def print_params(params, values, loss=None):
    table = tt.Texttable()
    table.header(["Parameter", "Value"])

    for param, value in zip(params, values):
        table.add_row([param.name, value])
    if loss is not None:
        table.add_row(["Loss value:", loss])
    print(table.draw())


def print_gradient(params, values, gradient, loss=None):
    table = tt.Texttable()
    table.header(["Parameter", "Value", "Gradient"])
    for param, value, grad in zip(params, values, gradient):
        table.add_row([param.name, value, grad])
    if loss is not None:
        table.add_row(["Loss value:", loss, "|"])
    print(table.draw())



#  Copyright (c) 2024 zfit

import pytest


def test_moprhing_sum():
    import numpy as np
    import zfit
    import zfit.z.numpy as znp  # numpy-like backend interface

    normal_np = np.random.normal(loc=2.0, scale=1.3, size=10000)
    obs = zfit.Space("x", limits=(-10, 10))
    true_mu = 1.0
    mu = zfit.Parameter("mu", true_mu, -4, 6)
    true_sigma = 1.0
    sigma = zfit.Parameter("sigma", true_sigma, 0.1, 10)
    model_nobin = zfit.pdf.Gauss(mu, sigma, obs)
    data_nobin = zfit.Data.from_numpy(obs, normal_np)
    minimizer = zfit.minimize.Minuit()
    # make binned
    nbins = 50
    data = data_nobin.to_binned(nbins)
    model = model_nobin.to_binned(data.space)
    obs_binned = data.space
    sig_true = 10_000
    sig_yield = zfit.Parameter("sig_yield", sig_true, 0, 100_000)
    template_hist = model.sample(
        25_000
    ).to_hist()  # let's assume we have 25_000 MC events in our template
    histpdf = zfit.pdf.HistogramPDF(template_hist, extended=sig_yield)
    sig_pdf = zfit.pdf.BinwiseScaleModifier(
        histpdf, modifiers=True
    )  # or we could give a list of parameters matching each bin
    modifiers = sig_pdf.params
    bkg_hist = zfit.Data.from_numpy(
        obs=obs, array=np.random.exponential(scale=20, size=100_000) - 10
    ).to_binned(obs_binned)
    bkg_hist_m1 = zfit.Data.from_numpy(
        obs=obs, array=np.random.exponential(scale=35, size=100_000) - 10
    ).to_binned(obs_binned)
    bkg_hist_m05 = zfit.Data.from_numpy(
        obs=obs, array=np.random.exponential(scale=26, size=100_000) - 10
    ).to_binned(obs_binned)
    bkg_hist_p1 = zfit.Data.from_numpy(
        obs=obs, array=np.random.exponential(scale=17, size=100_000) - 10
    ).to_binned(obs_binned)

    bkg_hists = {-1: bkg_hist_m1, -0.5: bkg_hist_m05, 0: bkg_hist, 1: bkg_hist_p1}
    bkg_histpdfs = {k: zfit.pdf.HistogramPDF(v) for k, v in bkg_hists.items()}
    alpha = zfit.Parameter("alpha", 0, -3, 3)
    bkg_true = 45_000
    bkg_yield = zfit.Parameter("bkg_yield", bkg_true)
    bkg_pdf = zfit.pdf.SplineMorphingPDF(alpha, bkg_histpdfs, extended=bkg_yield)

    model = zfit.pdf.BinnedSumPDF([sig_pdf, bkg_pdf])
    with zfit.param.set_values([alpha], [0.1]):
        data = model.sample()
    uncertainties = 1 / znp.maximum(template_hist.counts() ** 0.5, 1)
    modifier_constraints = zfit.constraint.GaussianConstraint(
        params=list(modifiers.values()),
        observation=np.ones(len(modifiers)),
        uncertainty=uncertainties,
    )
    alpha_constraint = zfit.constraint.GaussianConstraint(alpha, 0, 1)
    loss_binned = zfit.loss.ExtendedBinnedNLL(
        model, data, constraints=[modifier_constraints, alpha_constraint]
    )
    bkg_yield.set_value(bkg_true * 1.1)
    sig_yield.set_value(sig_true * 0.9)

    result = minimizer.minimize(loss_binned)
    assert result.valid
    params_to_test = [bkg_yield, alpha] + list(modifiers.values())[::7]

    result.hesse(name="hesse", params=params_to_test)
    result.errors(name="zfit", method="zfit_errors", params=params_to_test)
    result.errors(name="minos", method="minuit_minos", params=params_to_test)
    assert pytest.approx(sig_true, abs=3 * sig_true**0.5) == np.array(sig_yield)
    assert (
        pytest.approx(bkg_true, abs=3 * bkg_true**0.5)
        == result.params[bkg_yield]["value"]
    )
    assert pytest.approx(1, abs=0.003) == np.mean(list(modifiers.values()))
    for p in params_to_test:
        p = result.params[p]
        lower = p["zfit"]["lower"]
        upper = p["zfit"]["upper"]
        avg = (abs(upper) + abs(lower)) / 2
        assert pytest.approx(lower, rel=0.03) == p["minos"]["lower"]
        assert pytest.approx(upper, rel=0.03) == p["minos"]["upper"]
        assert pytest.approx(avg, rel=0.15) == p["hesse"]["error"]
